/// @file   DpmAdapter.cpp
/// @brief  DPM wrapper.
/// @author Alejandro Álvarez Ayllón <aalvarez@cern.ch>
#include <serrno.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <dmlite/common/errno.h>
#include <dmlite/cpp/utils/urls.h>
#include <dpm_api.h>
#include <map>
#include <vector>

#include "Adapter.h"
#include "DpmAdapter.h"
#include "FunctionWrapper.h"

using namespace dmlite;



DpmAdapterCatalog::DpmAdapterCatalog(unsigned retryLimit, bool hostDnIsRoot, std::string hostDn) throw (DmException):
  NsAdapterCatalog(retryLimit, hostDnIsRoot, hostDn)
{
  dpm_client_resetAuthorizationId();
}



DpmAdapterCatalog::~DpmAdapterCatalog()
{
  // Or it will be pointing to a freed area!
  dpm_client_resetAuthorizationId();
}



std::string DpmAdapterCatalog::getImplId() const throw ()
{
  return std::string("DpmAdapterCatalog");
}



void DpmAdapterCatalog::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  NsAdapterCatalog::setSecurityContext(ctx);
  
  // Root is a special case
  if (ctx->user.getUnsigned("uid") == 0) {
    FunctionWrapper<int> reset(dpm_client_resetAuthorizationId);
    reset();
  }
  else {
    // Call DPM API
    FunctionWrapper<int, uid_t, gid_t, const char*, char*>(dpm_client_setAuthorizationId,
                                                           ctx->user.getUnsigned("uid"),
                                                           ctx->groups[0].getUnsigned("gid"),
                                                           "GSI",
                                                           (char*)ctx->user.name.c_str())();

    if (this->nFqans_ > 0)
      FunctionWrapper<int, char*, char**, int>(dpm_client_setVOMS_data, (char*)ctx->groups[0].name.c_str(),
                                               (char**)this->fqans_, this->nFqans_)();
  }
}



void DpmAdapterCatalog::unlink(const std::string& path) throw (DmException)
{
  int                    nReplies;
  struct dpm_filestatus *statuses;
  
  std::string            absolute;

  if (path[0] == '/')
    absolute = path;
  else
    absolute = this->cwdPath_ + "/" + path;

  struct stat stat = NsAdapterCatalog::extendedStat(absolute, false).stat;
  if (S_ISLNK(stat.st_mode)) {
    NsAdapterCatalog::unlink(absolute);
  }
  else {
    const char *path_c = absolute.c_str();
    FunctionWrapper<int, int, char**, int*, dpm_filestatus**>(dpm_rm, 1, (char**)&path_c, &nReplies, &statuses)(this->retryLimit_);
    dpm_free_filest(nReplies, statuses);
  }
}



DpmAdapterPoolManager::DpmAdapterPoolManager(unsigned retryLimit,
                                             const std::string& passwd,
                                             bool useIp,
                                             unsigned life) throw (DmException):
  si_(NULL), retryLimit_(retryLimit), tokenPasswd_(passwd), tokenUseIp_(useIp),
  tokenLife_(life), userId_(""), fqans_(NULL), nFqans_(0)
{
  // Nothing
}



DpmAdapterPoolManager::~DpmAdapterPoolManager()
{
  if (this->fqans_ != NULL) {
    for (size_t i = 0; i < this->nFqans_; ++i)
      delete [] this->fqans_[i];
    delete [] this->fqans_;
  }
}



std::string DpmAdapterPoolManager::getImplId() const throw()
{
  return std::string("DpmAdapter");
}



void DpmAdapterPoolManager::setStackInstance(StackInstance* si) throw (DmException)
{
  // Nothing
  this->si_ = si;
}



void DpmAdapterPoolManager::setSecurityContext(const SecurityContext* ctx) throw (DmException)
{
  // String => const char*
  if (this->fqans_ != NULL) {
    for (size_t i = 0; i < this->nFqans_; ++i)
      delete [] this->fqans_[i];
    delete [] this->fqans_;
  }
  
  this->nFqans_ = ctx->groups.size();
  this->fqans_  = new char* [this->nFqans_];
  for (size_t i = 0; i < this->nFqans_; ++i) {
    this->fqans_[i] = new char [ctx->groups[i].name.length() + 1];
    strcpy(this->fqans_[i], ctx->groups[i].name.c_str());
  }

  // Call DPM API
  FunctionWrapper<int, uid_t, gid_t, const char*, char*>(dpm_client_setAuthorizationId,
                                                         ctx->user.getUnsigned("uid"),
                                                         ctx->groups[0].getUnsigned("gid"),
                                                         "GSI",
                                                         (char*)ctx->user.name.c_str())();

  if (!ctx->groups.empty()) {
    FunctionWrapper<int, char*, char**, int>(dpm_client_setVOMS_data,
                                             (char*)ctx->groups[0].name.c_str(),
                                             (char**)this->fqans_,
                                             ctx->groups.size())();
  }
  
  if (this->tokenUseIp_)
    this->userId_ = ctx->credentials.remoteAddress;
  else
    this->userId_ = ctx->credentials.clientName;
}



std::vector<Pool> DpmAdapterPoolManager::getPools(PoolAvailability availability) throw (DmException)
{
  struct dpm_pool* dpmPools = NULL;
  int              nPools;

  try {
    // Get all pools
    FunctionWrapper<int, int*, dpm_pool**>(dpm_getpools, &nPools, &dpmPools)(this->retryLimit_);
    
    std::vector<Pool> pools;
    Pool              pool;

    for (int i = 0; i < nPools; ++i) {
      pool.name = dpmPools[i].poolname;
      pool.type = "filesystem";
      pools.push_back(pool);
    }
    
    free(dpmPools);
    
    if (availability == kAny)
      return pools;
    
    // A pool is available if it has at least one fs available
    struct dpm_fs    *dpm_fs;
    std::vector<Pool> filtered;
  
    for (unsigned i = 0; i < pools.size(); ++i) {
      int  nFs;
      bool anyFsAvailable;

      if (dpm_getpoolfs((char*)pools[i].name.c_str(), &nFs, &dpm_fs) < 0)
        ThrowExceptionFromSerrno(serrno);

      anyFsAvailable = false;
      for (int j = 0; j < nFs && !anyFsAvailable; ++j) {
        
        switch (availability) {
          case kForWrite: case kForBoth:
            if (dpm_fs[j].status == 0)
              anyFsAvailable = true;
            break;
          default:
            if (dpm_fs[j].status != FS_DISABLED)
              anyFsAvailable = true;
        }

      }
      
      // Unless kNone, anyFsAvailable means a match
      if ((availability == kNone && !anyFsAvailable) ||
          (availability != kNone && anyFsAvailable))
        filtered.push_back(pools[i]);
      
      if (nFs > 0)
        free(dpm_fs);
    }
    
    return filtered;    
  }
  catch (...) {
    if (dpmPools != NULL)
      free(dpmPools);
    throw;
  }
}



Pool DpmAdapterPoolManager::getPool(const std::string& poolname) throw (DmException)
{
  std::vector<Pool> pools = this->getPools();
  
  for (unsigned i = 0; i < pools.size(); ++i) {
    if (poolname == pools[i].name)
      return pools[i];
  }
  
  throw DmException(DMLITE_NO_SUCH_POOL,
                    "Pool " + poolname + " not found");
}



void DpmAdapterPoolManager::newPool(const Pool&) throw (DmException)
{
  throw DmException(DMLITE_SYSERR(ENOSYS),
                    "DpmAdapterPoolManager::newPool not implemented");
}



void DpmAdapterPoolManager::updatePool(const Pool&) throw (DmException)
{
  throw DmException(DMLITE_SYSERR(ENOSYS),
                    "DpmAdapterPoolManager::updatePool not implemented");
}



void DpmAdapterPoolManager::deletePool(const Pool&) throw (DmException)
{
  throw DmException(DMLITE_SYSERR(ENOSYS),
                    "DpmAdapterPoolManager::deletePool not implemented");
}



Location DpmAdapterPoolManager::whereToRead(const std::string& path) throw (DmException)
{
  struct dpm_getfilereq     request;
  struct dpm_getfilestatus *statuses = NULL;
  int                       nReplies, wait;
  char                      r_token[CA_MAXDPMTOKENLEN + 1];
  std::string               rfn;

  request.from_surl  = (char*)path.c_str();
  request.s_token[0] = '\0';
  request.ret_policy = '\0';
  request.flags      = 0;

  if (si_->contains("lifetime"))
    request.lifetime = Extensible::anyToLong(si_->get("lifetime"));
  else
    request.lifetime = 0;

  if (si_->contains("f_type"))
    request.f_type = Extensible::anyToString(si_->get("f_type"))[0];
  else
    request.f_type = '\0';

  try {
    // Request
    char rfiostr[] = "rfio";
    char *protocols[] = {rfiostr};
    char description[] = "dmlite::adapter::whereToRead";

    FunctionWrapper<int, int, dpm_getfilereq*, int, char**, char*, time_t, char*, int*, dpm_getfilestatus**>
      (dpm_get, 1, &request, 1, protocols, description, 0, r_token, &nReplies, &statuses)(this->retryLimit_);

    if (nReplies < 1)
      throw DmException(DMLITE_NO_REPLICAS, "No replicas found for " + path);

    // Wait for answer
    wait = (statuses[0].status == DPM_QUEUED ||
            statuses[0].status == DPM_RUNNING ||
            statuses[0].status == DPM_ACTIVE);

    FunctionWrapper <int, char*, int, char**, int*, dpm_getfilestatus**> getReq(dpm_getstatus_getreq, r_token, 1,
                                                                                &request.from_surl, &nReplies, &statuses);
    while (wait) {
      dpm_free_gfilest(nReplies, statuses);
      getReq();
      if (!nReplies)
        throw DmException(DMLITE_NO_REPLICAS, "No replicas found for " + path);

      wait = (statuses[0].status == DPM_QUEUED ||
              statuses[0].status == DPM_RUNNING ||
              statuses[0].status == DPM_ACTIVE);
    }

    switch(statuses[0].status & 0xF000) {
      case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
        throw DmException(DMLITE_SYSERR(statuses[0].status & 0x0FFF),
                          "The DPM get request failed (%s)",
                          statuses[0].errstring?statuses[0].errstring:"No error string returned from DPM");
    }

    rfn = statuses[0].turl;
    dpm_free_gfilest(nReplies, statuses);
    
    // Build the answer
    Url rloc(rfn);
    rloc.path = Url::normalizePath(rloc.path);
    
    Chunk single;
    
    single.url.domain = rloc.domain;
    single.url.path   = rloc.path;
    
    single.offset = 0;
    single.size   = this->si_->getCatalog()->extendedStat(path).stat.st_size;
    
    single.url.query["token"] = dmlite::generateToken(this->userId_, rloc.path,
                                                      this->tokenPasswd_, this->tokenLife_);
    
    return Location(1, single);          
  }
  catch (...) {
    // On exceptions, free first!
    if (statuses != NULL)
      dpm_free_gfilest(nReplies, statuses);
    throw;
  }
}



Location DpmAdapterPoolManager::whereToRead(ino_t) throw (DmException)
{
  throw DmException(DMLITE_SYSERR(ENOSYS),
                    "DpmAdapter: Access by inode not supported");
}



Location DpmAdapterPoolManager::whereToWrite(const std::string& path) throw (DmException)
{
  struct dpm_putfilereq     reqfile;
  struct dpm_putfilestatus *statuses = NULL;
  int                       nReplies, wait;
  char                      token[CA_MAXDPMTOKENLEN + 1];
  std::string               spaceToken, userTokenDescription;

  mode_t createMode = 0664;
  Acl    acl;

  reqfile.to_surl        = (char*)path.c_str();
  reqfile.s_token[0]     = '\0';

  if (this->si_->contains("f_type"))
    reqfile.f_type = Extensible::anyToString(this->si_->get("f_type"))[0];
  else
    reqfile.f_type = 'P';

  if (this->si_->contains("lifetime"))
    reqfile.lifetime = Extensible::anyToLong(this->si_->get("lifetime"));
  else
    reqfile.lifetime = 0;

  if (this->si_->contains("requested_size"))
    reqfile.requested_size = Extensible::anyToLong(this->si_->get("requested_size"));
  else
    reqfile.requested_size = 0;

  if (this->si_->contains("f_lifetime"))
    reqfile.f_lifetime = Extensible::anyToLong(this->si_->get("f_lifetime"));
  else
    reqfile.f_lifetime = 0;

  if (this->si_->contains("ret_policy"))
    reqfile.ret_policy = Extensible::anyToString(this->si_->get("ret_policy"))[0];
  else
    reqfile.ret_policy     = '\0';

  if (this->si_->contains("ac_latency"))
    reqfile.ac_latency = Extensible::anyToString(this->si_->get("ac_latency"))[0];
  else
    reqfile.ac_latency     = '\0';
  
  if (this->si_->contains("overwrite") && Extensible::anyToBoolean(this->si_->get("overwrite"))) {
    try {
      ExtendedStat xstat = this->si_->getCatalog()->extendedStat(path);
      acl        = xstat.acl;
      createMode = xstat.stat.st_mode;
      this->si_->getCatalog()->unlink(path);
    }
    catch (DmException& e) {
      if (e.code() != ENOENT) throw;
    }
  }

  if (this->si_->contains("SpaceToken"))
    spaceToken           = Extensible::anyToString(this->si_->get("SpaceToken"));
  else if (this->si_->contains("UserSpaceTokenDescription"))
    userTokenDescription = Extensible::anyToString(this->si_->get("UserSpaceTokenDescription"));

  if (!spaceToken.empty()) {
    strncpy(reqfile.s_token, spaceToken.c_str(), sizeof(reqfile.s_token));
  }
  else if (!userTokenDescription.empty()) {
    char **space_ids = NULL;

    FunctionWrapper<int, const char*, int*, char***>
      (dpm_getspacetoken, userTokenDescription.c_str(), &nReplies, &space_ids)(this->retryLimit_);

    if (nReplies > 0)
      strncpy(reqfile.s_token, space_ids[0], sizeof(reqfile.s_token));

    for(int i = 0; i < nReplies; ++i)
      free(space_ids[i]);
    free(space_ids);

    if (nReplies == 0)
      throw DmException(EINVAL, "Could not get the space token from the token description %s",
                        userTokenDescription.c_str());
  }

  this->si_->getCatalog()->create(path, createMode);
  if (!acl.empty())
    this->si_->getCatalog()->setAcl(path, acl);

  try {
    char rfiostr[] = "rfio";
    char *protocols[] = { rfiostr };
    char description[] = "dmlite::adapter::whereToWrite::put";
    FunctionWrapper<int, int, dpm_putfilereq*, int, char**, char*, int, time_t, char*, int*, dpm_putfilestatus**>
      (dpm_put, 1, &reqfile, 1, protocols, description, 1, 0, token, &nReplies, &statuses)(this->retryLimit_);

    wait = (statuses[0].status == DPM_QUEUED  ||
            statuses[0].status == DPM_RUNNING ||
            statuses[0].status == DPM_ACTIVE);

    FunctionWrapper <int, char*, int, char**, int*, dpm_putfilestatus**> putReq(dpm_getstatus_putreq, token, 1,
                                                                                &reqfile.to_surl, &nReplies, &statuses);
    while (wait) {
      dpm_free_pfilest(nReplies, statuses);
      putReq();
      if (!nReplies)
        throw DmException(DMLITE_SYSERR(serrno),
                          "Didn't get a destination from DPM");

      wait = (statuses[0].status == DPM_QUEUED  ||
                  statuses[0].status == DPM_RUNNING ||
                  statuses[0].status == DPM_ACTIVE);
    }
    
    switch(statuses[0].status & 0xF000) {
      case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
        throw DmException(serrno,
                          "The DPM put request failed (%s)",
                          statuses[0].errstring?statuses[0].errstring:"No error string returned from DPM");
    }

    Url rloc(statuses[0].turl);
    dpm_free_pfilest(nReplies, statuses);
    
    rloc.path = Url::normalizePath(rloc.path);
    
    Chunk single;
    
    single.url.domain = rloc.domain;
    single.url.path   = rloc.path;
    
    single.offset = 0;
    single.size   = 0;
    
    single.url.query["sfn"]      = path;
    single.url.query["dpmtoken"] = std::string(token);
    single.url.query["token"]    = dmlite::generateToken(this->userId_, rloc.path,
                                                         this->tokenPasswd_, this->tokenLife_, true);
    
    return Location(1, single);
  }
  catch (...) {
    // On exceptions, free first!
    if (statuses != NULL)
      dpm_free_pfilest(nReplies, statuses);
    throw;
  }
}



void DpmAdapterPoolManager::cancelWrite(const Location& loc) throw (DmException)
{
  FunctionWrapper<int, char*>(dpm_abortreq, (char*)loc[0].url.query.getString("dpmtoken").c_str())();
  // We do not wrap this, since dpm_abortreq may have removed the file already
  dpns_unlink(loc[0].url.query.getString("sfn").c_str());
}
